@@ -11,7 +11,7 @@ Set up the cloud configuration at ``/etc/salt/cloud.providers`` or
 .. code-block:: yaml
 
     my-joyent-config:
-      provider: joyent
+      driver: joyent
       # The Joyent login user
       user: fred
       # The Joyent user's password
@@ -48,14 +48,11 @@ included:
 
 :depends: PyCrypto
 '''
-from __future__ import absolute_import
-# pylint: disable=E0102
-
-# The import section is mostly libcloud boilerplate
+# pylint: disable=invalid-name,function-redefined
 
 # Import python libs
+from __future__ import absolute_import
 import os
-import salt.ext.six.moves.http_client  # pylint: disable=E0611
 import json
 import logging
 import base64
@@ -67,27 +64,25 @@ from Crypto.Hash import SHA256
 from Crypto.PublicKey import RSA
 from Crypto.Signature import PKCS1_v1_5
 
-# Import generic libcloud functions
-from salt.cloud.libcloudfuncs import *   # pylint: disable=W0614,W0401
-
 # Import salt libs
 import salt.ext.six as six
+from salt.ext.six.moves import http_client  # pylint: disable=import-error,no-name-in-module
 import salt.utils.http
 import salt.utils.cloud
 import salt.config as config
-from salt.utils import namespaced_function
 from salt.utils.cloud import is_public_ip
+from salt.cloud.libcloudfuncs import node_state
 from salt.exceptions import (
     SaltCloudSystemExit,
     SaltCloudExecutionFailure,
-    SaltCloudExecutionTimeout
+    SaltCloudExecutionTimeout,
+    SaltCloudNotFound,
 )
 
 # Get logging started
 log = logging.getLogger(__name__)
 
-# namespace libcloudfuncs
-get_salt_interface = namespaced_function(get_salt_interface, globals())
+__virtualname__ = 'joyent'
 
 JOYENT_API_HOST_SUFFIX = '.api.joyentcloud.com'
 JOYENT_API_VERSION = '~7.2'
@@ -106,25 +101,22 @@ DEFAULT_LOCATION = 'us-east-1'
 POLL_ALL_LOCATIONS = True
 
 VALID_RESPONSE_CODES = [
-    salt.ext.six.moves.http_client.OK,
-    salt.ext.six.moves.http_client.ACCEPTED,
-    salt.ext.six.moves.http_client.CREATED,
-    salt.ext.six.moves.http_client.NO_CONTENT
+    http_client.OK,
+    http_client.ACCEPTED,
+    http_client.CREATED,
+    http_client.NO_CONTENT
 ]
 
 
-# Only load in this module is the JOYENT configurations are in place
+# Only load in this module if the Joyent configurations are in place
 def __virtual__():
     '''
-    Set up the libcloud functions and check for JOYENT configs
+    Check for Joyent configs
     '''
     if get_configured_provider() is False:
         return False
 
-    global script
-    conn = None
-    script = namespaced_function(script, globals(), (conn,))
-    return True
+    return __virtualname__
 
 
 def get_configured_provider():
@@ -133,7 +125,7 @@ def get_configured_provider():
     '''
     return config.is_provider_configured(
         __opts__,
-        __active_provider_name__ or 'joyent',
+        __active_provider_name__ or __virtualname__,
         ('user', 'password')
     )
 
@@ -204,7 +196,7 @@ def query_instance(vm_=None, call=None):
 
         if isinstance(data, dict) and 'error' in data:
             log.warn(
-                'There was an error in the query. {0}'.format(data['error'])
+                'There was an error in the query {0}'.format(data['error'])  # pylint: disable=E1126
             )
             # Trigger a failure in the wait for IP function
             return False
@@ -248,6 +240,20 @@ def create(vm_):
 
         salt-cloud -p profile_name vm_name
     '''
+    try:
+        # Check for required profile parameters before sending any API calls.
+        if vm_['profile'] and config.is_profile_configured(__opts__,
+                                                           __active_provider_name__ or 'joyent',
+                                                           vm_['profile']) is False:
+            return False
+    except AttributeError:
+        pass
+
+    # Since using "provider: <provider-engine>" is deprecated, alias provider
+    # to use driver: "driver: <provider-engine>"
+    if 'provider' in vm_:
+        vm_['driver'] = vm_.pop('provider')
+
     key_filename = config.get_cloud_config_value(
         'private_key', vm_, __opts__, search_global=False, default=None
     )
@@ -259,7 +265,7 @@ def create(vm_):
         {
             'name': vm_['name'],
             'profile': vm_['profile'],
-            'provider': vm_['provider'],
+            'provider': vm_['driver'],
         },
         transport=__opts__['transport']
     )
@@ -271,7 +277,7 @@ def create(vm_):
         )
     )
 
-    ## added . for fqdn hostnames
+    # added . for fqdn hostnames
     salt.utils.cloud.check_name(vm_['name'], 'a-zA-Z0-9-.')
     kwargs = {
         'name': vm_['name'],
@@ -318,7 +324,7 @@ def create(vm_):
         {
             'name': vm_['name'],
             'profile': vm_['profile'],
-            'provider': vm_['provider'],
+            'provider': vm_['driver'],
         },
         transport=__opts__['transport']
     )
@@ -642,7 +648,7 @@ def joyent_node_state(id_):
     values for consistency
 
     :param id_: joyent state value
-    :return: libcloudfuncs state value
+    :return: state value
     '''
     states = {'running': 0,
               'stopped': 2,
@@ -685,10 +691,14 @@ def reformat_node(item=None, full=False):
             item[key] = None
 
     # remove all the extra key value pairs to provide a brief listing
+    to_del = []
     if not full:
-        for key in item.keys():  # iterate over a copy of the keys
+        for key in six.iterkeys(item):  # iterate over a copy of the keys
             if key not in desired_keys:
-                del item[key]
+                to_del.append(key)
+
+    for key in to_del:
+        del item[key]
 
     if 'state' in item:
         item['state'] = joyent_node_state(item['state'])
@@ -1069,7 +1079,7 @@ def query(action=None,
         text=True,
         status=True,
         headers=True,
-        verify=verify_ssl,
+        verify_ssl=verify_ssl,
         opts=__opts__,
     )
     log.debug(
