@@ -1,4 +1,4 @@
-## -*- coding: utf-8 -*-
+# -*- coding: utf-8 -*-
 '''
 The networking module for Debian based distros
 
@@ -8,17 +8,19 @@ References:
 '''
 
 # Import python libs
+from __future__ import absolute_import
 import functools
 import logging
 import os.path
 import os
 import re
-import StringIO
 import time
 
 # Import third party libs
 import jinja2
 import jinja2.exceptions
+import salt.ext.six as six
+from salt.ext.six.moves import StringIO  # pylint: disable=import-error,no-name-in-module
 
 # Import salt libs
 import salt.utils
@@ -123,24 +125,26 @@ _DEB_CONFIG_PPPOE_OPTS = {
     'persist': 'persist',
     'mtu': 'mtu',
     'noaccomp': 'noaccomp',
+    'linkname': 'linkname',
 }
 
 _DEB_ROUTES_FILE = '/etc/network/routes'
 _DEB_NETWORK_FILE = '/etc/network/interfaces'
-_DEB_NETWORK_DIR = '/etc/network/interfaces.d'
-_DEB_NETWORK_UP_DIR = '/etc/network/if-up.d'
-_DEB_NETWORK_DOWN_DIR = '/etc/network/if-down.d'
-_DEB_NETWORK_CONF_FILES = '/etc/modprobe.d'
+_DEB_NETWORK_DIR = '/etc/network/interfaces.d/'
+_DEB_NETWORK_UP_DIR = '/etc/network/if-up.d/'
+_DEB_NETWORK_DOWN_DIR = '/etc/network/if-down.d/'
+_DEB_NETWORK_CONF_FILES = '/etc/modprobe.d/'
 _DEB_NETWORKING_FILE = '/etc/default/networking'
 _DEB_HOSTNAME_FILE = '/etc/hostname'
 _DEB_RESOLV_FILE = '/etc/resolv.conf'
-_DEB_PPP_DIR = '/etc/ppp/peers'
+_DEB_PPP_DIR = '/etc/ppp/peers/'
 
 _CONFIG_TRUE = ['yes', 'on', 'true', '1', True]
 _CONFIG_FALSE = ['no', 'off', 'false', '0', False]
 _IFACE_TYPES = [
     'eth', 'bond', 'alias', 'clone',
-    'ipsec', 'dialup', 'bridge', 'slave', 'vlan', 'pppoe',
+    'ipsec', 'dialup', 'bridge', 'slave',
+    'vlan', 'pppoe', 'source',
 ]
 
 
@@ -210,11 +214,11 @@ def _raise_error_routes(iface, option, expected):
 
 def _read_file(path):
     '''
-    Reads and returns the contents of a file
+    Reads and returns the contents of a text file
     '''
     try:
         with salt.utils.flopen(path, 'rb') as contents:
-            return contents.readlines()
+            return [salt.utils.to_str(line) for line in contents.readlines()]
     except (OSError, IOError):
         return ''
 
@@ -246,7 +250,10 @@ def _parse_hostname():
     Parse /etc/hostname and return hostname
     '''
     contents = _read_file(_DEB_HOSTNAME_FILE)
-    return contents[0].split('\n')[0]
+    if contents:
+        return contents[0].split('\n')[0]
+    else:
+        return ''
 
 
 def _parse_current_network_settings():
@@ -331,9 +338,9 @@ def __ipv4_netmask(value):
 
 def __ipv6_netmask(value):
     '''validate an IPv6 integer netmask'''
-    valid, errmsg = False, 'IPv6 netmask (0->127)'
+    valid, errmsg = False, 'IPv6 netmask (0->128)'
     valid, value, _ = __int(value)
-    valid = (valid and 0 <= value <= 127)
+    valid = (valid and 0 <= value <= 128)
     return (valid, value, errmsg)
 
 
@@ -389,13 +396,13 @@ SALT_ATTR_TO_DEBIAN_ATTR_MAP = {
 
 
 DEBIAN_ATTR_TO_SALT_ATTR_MAP = dict(
-    (v, k) for (k, v) in SALT_ATTR_TO_DEBIAN_ATTR_MAP.items())
+    (v, k) for (k, v) in six.iteritems(SALT_ATTR_TO_DEBIAN_ATTR_MAP))
 
-#TODO
+# TODO
 DEBIAN_ATTR_TO_SALT_ATTR_MAP['address'] = 'address'
 DEBIAN_ATTR_TO_SALT_ATTR_MAP['hwaddress'] = 'hwaddress'
 
-IPV4_VALID_PROTO = ['bootp', 'dhcp', 'static', 'manual', 'loopback']
+IPV4_VALID_PROTO = ['bootp', 'dhcp', 'static', 'manual', 'loopback', 'ppp']
 
 IPV4_ATTR_MAP = {
     'proto': __within(IPV4_VALID_PROTO, dtype=str),
@@ -420,7 +427,7 @@ IPV4_ATTR_MAP = {
     'server':  __ipv4_quad,
     'hwaddr':  __mac,
     # tunnel
-    'mode':  __within(['gre', 'GRE', 'ipip', 'IPIP'], dtype=str),
+    'mode':  __within(['gre', 'GRE', 'ipip', 'IPIP', '802.3ad'], dtype=str),
     'endpoint':  __ipv4_quad,
     'dstaddr':  __ipv4_quad,
     'local':  __ipv4_quad,
@@ -518,8 +525,8 @@ def _parse_interfaces(interface_files=None):
     if interface_files is None:
         interface_files = []
         # Add this later.
-        # if os.path.exists('/etc/network/interfaces.d/'):
-        #    interface_files += os.listdir('/etc/network/interfaces.d/')
+        if os.path.exists(_DEB_NETWORK_DIR):
+            interface_files += ['{0}/{1}'.format(_DEB_NETWORK_DIR, dir) for dir in os.listdir(_DEB_NETWORK_DIR)]
 
         if os.path.isfile(_DEB_NETWORK_FILE):
             interface_files.insert(0, _DEB_NETWORK_FILE)
@@ -564,6 +571,7 @@ def _parse_interfaces(interface_files=None):
 
                     iface_dict['addrfam'] = addrfam
                     iface_dict['proto'] = method
+                    iface_dict['filename'] = interface_file
 
                 # Parse the detail clauses.
                 elif line[0].isspace():
@@ -574,11 +582,10 @@ def _parse_interfaces(interface_files=None):
 
                     attr, valuestr = line.rstrip().split(None, 1)
                     if _attrmaps_contain_attr(attr):
-                        _attr = DEBIAN_ATTR_TO_SALT_ATTR_MAP.get(attr, attr)
-                        if '-' in _attr:
-                            attrname = _attr.replace('-', '_')
+                        if '-' in attr:
+                            attrname = attr.replace('-', '_')
                         else:
-                            attrname = _attr
+                            attrname = attr
                         (valid, value, errmsg) = _validate_interface_option(
                             attr, valuestr, addrfam)
                         iface_dict[attrname] = value
@@ -589,13 +596,13 @@ def _parse_interfaces(interface_files=None):
                         iface_dict['ethtool'][attr] = valuestr
 
                     elif attr.startswith('bond'):
-                        opt = attr.split('_', 1)[1]
+                        opt = re.split(r'[_-]', attr, maxsplit=1)[1]
                         if 'bonding' not in iface_dict:
                             iface_dict['bonding'] = salt.utils.odict.OrderedDict()
                         iface_dict['bonding'][opt] = valuestr
 
                     elif attr.startswith('bridge'):
-                        opt = attr.split('_', 1)[1]
+                        opt = re.split(r'[_-]', attr, maxsplit=1)[1]
                         if 'bridging' not in iface_dict:
                             iface_dict['bridging'] = salt.utils.odict.OrderedDict()
                         iface_dict['bridging'][opt] = valuestr
@@ -620,9 +627,26 @@ def _parse_interfaces(interface_files=None):
                             adapters[word] = salt.utils.odict.OrderedDict()
                         adapters[word]['hotplug'] = True
 
+                elif line.startswith('source'):
+                    if 'source' not in adapters:
+                        adapters['source'] = salt.utils.odict.OrderedDict()
+
+                    # Create item in dict, if not already there
+                    if 'data' not in adapters['source']:
+                        adapters['source']['data'] = salt.utils.odict.OrderedDict()
+                        adapters['source']['data']['sources'] = []
+                    adapters['source']['data']['sources'].append(line.split()[1])
+
     # Return a sorted list of the keys for bond, bridge and ethtool options to
     # ensure a consistent order
     for iface_name in adapters:
+        if iface_name == 'source':
+            continue
+        if 'data' not in adapters[iface_name]:
+            msg = 'Interface file malformed for interface: {0}.'.format(iface_name)
+            log.error(msg)
+            adapters.pop(iface_name)
+            continue
         for opt in ['ethtool', 'bonding', 'bridging']:
             if 'inet' in adapters[iface_name]['data']:
                 if opt in adapters[iface_name]['data']['inet']:
@@ -689,11 +713,8 @@ def _parse_ethtool_pppoe_opts(opts, iface):
         if opt in opts:
             config[opt] = opts[opt]
 
-    if 'provider' in opts:
-        if opts['provider']:
-            pass
-        else:
-            _raise_error_iface(iface, 'provider', _CONFIG_TRUE + _CONFIG_FALSE)
+    if 'provider' in opts and not opts['provider']:
+        _raise_error_iface(iface, 'provider', _CONFIG_TRUE + _CONFIG_FALSE)
 
     valid = _CONFIG_TRUE + _CONFIG_FALSE
     for option in ('noipdefault', 'usepeerdns', 'defaultroute', 'hide-password', 'noauth', 'persist', 'noaccomp'):
@@ -858,9 +879,9 @@ def _parse_settings_bond_1(opts, iface, bond_def):
 
     if 'use_carrier' in opts:
         if opts['use_carrier'] in _CONFIG_TRUE:
-            bond.update({'use_carrier': 'on'})
+            bond.update({'use_carrier': '1'})
         elif opts['use_carrier'] in _CONFIG_FALSE:
-            bond.update({'use_carrier': 'off'})
+            bond.update({'use_carrier': '0'})
         else:
             valid = _CONFIG_TRUE + _CONFIG_FALSE
             _raise_error_iface(iface, 'use_carrier', valid)
@@ -941,9 +962,9 @@ def _parse_settings_bond_3(opts, iface, bond_def):
 
     if 'use_carrier' in opts:
         if opts['use_carrier'] in _CONFIG_TRUE:
-            bond.update({'use_carrier': 'on'})
+            bond.update({'use_carrier': '1'})
         elif opts['use_carrier'] in _CONFIG_FALSE:
-            bond.update({'use_carrier': 'off'})
+            bond.update({'use_carrier': '0'})
         else:
             valid = _CONFIG_TRUE + _CONFIG_FALSE
             _raise_error_iface(iface, 'use_carrier', valid)
@@ -985,9 +1006,9 @@ def _parse_settings_bond_4(opts, iface, bond_def):
 
     if 'use_carrier' in opts:
         if opts['use_carrier'] in _CONFIG_TRUE:
-            bond.update({'use_carrier': 'on'})
+            bond.update({'use_carrier': '1'})
         elif opts['use_carrier'] in _CONFIG_FALSE:
-            bond.update({'use_carrier': 'off'})
+            bond.update({'use_carrier': '0'})
         else:
             valid = _CONFIG_TRUE + _CONFIG_FALSE
             _raise_error_iface(iface, 'use_carrier', valid)
@@ -1028,9 +1049,9 @@ def _parse_settings_bond_5(opts, iface, bond_def):
 
     if 'use_carrier' in opts:
         if opts['use_carrier'] in _CONFIG_TRUE:
-            bond.update({'use_carrier': 'on'})
+            bond.update({'use_carrier': '1'})
         elif opts['use_carrier'] in _CONFIG_FALSE:
-            bond.update({'use_carrier': 'off'})
+            bond.update({'use_carrier': '0'})
         else:
             valid = _CONFIG_TRUE + _CONFIG_FALSE
             _raise_error_iface(iface, 'use_carrier', valid)
@@ -1064,9 +1085,9 @@ def _parse_settings_bond_6(opts, iface, bond_def):
 
     if 'use_carrier' in opts:
         if opts['use_carrier'] in _CONFIG_TRUE:
-            bond.update({'use_carrier': 'on'})
+            bond.update({'use_carrier': '1'})
         elif opts['use_carrier'] in _CONFIG_FALSE:
-            bond.update({'use_carrier': 'off'})
+            bond.update({'use_carrier': '0'})
         else:
             valid = _CONFIG_TRUE + _CONFIG_FALSE
             _raise_error_iface(iface, 'use_carrier', valid)
@@ -1184,12 +1205,14 @@ def _parse_settings_eth(opts, iface_type, enabled, iface):
     if iface_type == 'bridge':
         bridging = _parse_bridge_opts(opts, iface)
         if bridging:
+            opts.pop('mode', None)
             iface_data['inet']['bridging'] = bridging
             iface_data['inet']['bridging_keys'] = sorted(bridging)
 
     elif iface_type == 'bond':
         bonding = _parse_settings_bond(opts, iface)
         if bonding:
+            opts.pop('mode', None)
             iface_data['inet']['bonding'] = bonding
             iface_data['inet']['bonding']['slaves'] = opts['slaves']
             iface_data['inet']['bonding_keys'] = sorted(bonding)
@@ -1250,14 +1273,31 @@ def _parse_settings_eth(opts, iface_type, enabled, iface):
     return adapters
 
 
+def _parse_settings_source(opts, iface_type, enabled, iface):
+    '''
+    Filters given options and outputs valid settings for a
+    network interface.
+    '''
+    adapters = salt.utils.odict.OrderedDict()
+    adapters[iface] = salt.utils.odict.OrderedDict()
+
+    adapters[iface]['type'] = iface_type
+
+    adapters[iface]['data'] = salt.utils.odict.OrderedDict()
+    iface_data = adapters[iface]['data']
+    iface_data['sources'] = [opts['source']]
+
+    return adapters
+
+
 def _parse_network_settings(opts, current):
     '''
     Filters given options and outputs valid settings for
     the global network settings file.
     '''
     # Normalize keys
-    opts = dict((k.lower(), v) for (k, v) in opts.iteritems())
-    current = dict((k.lower(), v) for (k, v) in current.iteritems())
+    opts = dict((k.lower(), v) for (k, v) in six.iteritems(opts))
+    current = dict((k.lower(), v) for (k, v) in six.iteritems(current))
     result = {}
 
     valid = _CONFIG_TRUE + _CONFIG_FALSE
@@ -1299,7 +1339,7 @@ def _parse_routes(iface, opts):
     the route settings file.
     '''
     # Normalize keys
-    opts = dict((k.lower(), v) for (k, v) in opts.iteritems())
+    opts = dict((k.lower(), v) for (k, v) in six.iteritems(opts))
     result = {}
     if 'routes' not in opts:
         _raise_error_routes(iface, 'routes', 'List of routes')
@@ -1362,7 +1402,7 @@ def _read_temp(data):
     '''
     Return what would be written to disk
     '''
-    tout = StringIO.StringIO()
+    tout = StringIO()
     tout.write(data)
     tout.seek(0)
     output = tout.readlines()
@@ -1386,12 +1426,13 @@ def _read_temp_ifaces(iface, data):
     return [item + '\n' for item in ifcfg.split('\n')]
 
 
-def _write_file_ifaces(iface, data):
+def _write_file_ifaces(iface, data, **settings):
     '''
     Writes a file to disk
     '''
     try:
-        template = JINJA.get_template('debian_eth.jinja')
+        eth_template = JINJA.get_template('debian_eth.jinja')
+        source_template = JINJA.get_template('debian_source.jinja')
     except jinja2.exceptions.TemplateNotFound:
         log.error('Could not load template debian_eth.jinja')
         return ''
@@ -1405,23 +1446,41 @@ def _write_file_ifaces(iface, data):
     for adapter in adapters:
         if 'type' in adapters[adapter] and adapters[adapter]['type'] == 'slave':
             # Override values so the interfaces file is correct
-            adapters[adapter]['enabled'] = False
             adapters[adapter]['data']['inet']['addrfam'] = 'inet'
             adapters[adapter]['data']['inet']['proto'] = 'manual'
+            adapters[adapter]['data']['inet']['master'] = adapters[adapter]['master']
 
-        tmp = template.render({'name': adapter, 'data': adapters[adapter]})
-        ifcfg = tmp + ifcfg
+        if 'type' in adapters[adapter] and adapters[adapter]['type'] == 'source':
+            tmp = source_template.render({'name': adapter, 'data': adapters[adapter]})
+        else:
+            tmp = eth_template.render({'name': adapter, 'data': adapters[adapter]})
+        ifcfg = ifcfg + tmp
         if adapter == iface:
             saved_ifcfg = tmp
 
-    filename = _DEB_NETWORK_FILE
+    _SEPERATE_FILE = False
+    if 'filename' in settings:
+        if not settings['filename'].startswith('/'):
+            filename = '{0}/{1}'.format(_DEB_NETWORK_DIR, settings['filename'])
+        else:
+            filename = settings['filename']
+        _SEPERATE_FILE = True
+    else:
+        if 'filename' in adapters[adapter]['data']:
+            filename = adapters[adapter]['data']
+        else:
+            filename = _DEB_NETWORK_FILE
+
     if not os.path.exists(os.path.dirname(filename)):
         msg = '{0} cannot be written.'
         msg = msg.format(os.path.dirname(filename))
         log.error(msg)
         raise AttributeError(msg)
     with salt.utils.flopen(filename, 'w') as fout:
-        fout.write(ifcfg)
+        if _SEPERATE_FILE:
+            fout.write(saved_ifcfg)
+        else:
+            fout.write(ifcfg)
 
     # Return as a array so the difflib works
     return saved_ifcfg.split('\n')
@@ -1520,6 +1579,9 @@ def build_interface(iface, iface_type, enabled, **settings):
     if iface_type not in _IFACE_TYPES:
         _raise_error_iface(iface, iface_type, _IFACE_TYPES)
 
+    if 'proto' not in settings:
+        settings['proto'] = 'static'
+
     if iface_type == 'slave':
         settings['slave'] = 'yes'
         if 'master' not in settings:
@@ -1529,6 +1591,7 @@ def build_interface(iface, iface_type, enabled, **settings):
 
     elif iface_type == 'vlan':
         settings['vlan'] = 'yes'
+        __salt__['pkg.install']('vlan')
 
     elif iface_type == 'pppoe':
         settings['pppoe'] = 'yes'
@@ -1551,10 +1614,13 @@ def build_interface(iface, iface_type, enabled, **settings):
     if iface_type in ['eth', 'bond', 'bridge', 'slave', 'vlan', 'pppoe']:
         opts = _parse_settings_eth(settings, iface_type, enabled, iface)
 
+    if iface_type in ['source']:
+        opts = _parse_settings_source(settings, iface_type, enabled, iface)
+
     if 'test' in settings and settings['test']:
         return _read_temp_ifaces(iface, opts[iface])
 
-    ifcfg = _write_file_ifaces(iface, opts[iface])
+    ifcfg = _write_file_ifaces(iface, opts[iface], **settings)
 
     if iface_type == 'pppoe':
         _write_file_ppp_ifaces(iface, opts[iface])
@@ -1606,10 +1672,11 @@ def down(iface, iface_type):
 
     .. code-block:: bash
 
-        salt '*' ip.down eth0
+        salt '*' ip.down eth0 eth
     '''
     # Slave devices are controlled by the master.
-    if iface_type not in ['slave']:
+    # Source 'interfaces' aren't brought down.
+    if iface_type not in ['slave', 'source']:
         return __salt__['cmd.run']('ifdown {0}'.format(iface))
     return None
 
@@ -1642,7 +1709,10 @@ def get_interface(iface):
     adapters = _parse_interfaces()
     if iface in adapters:
         try:
-            template = JINJA.get_template('debian_eth.jinja')
+            if iface == 'source':
+                template = JINJA.get_template('debian_source.jinja')
+            else:
+                template = JINJA.get_template('debian_eth.jinja')
         except jinja2.exceptions.TemplateNotFound:
             log.error('Could not load template debian_eth.jinja')
             return ''
@@ -1663,10 +1733,11 @@ def up(iface, iface_type):  # pylint: disable=C0103
 
     .. code-block:: bash
 
-        salt '*' ip.up eth0
+        salt '*' ip.up eth0 eth
     '''
     # Slave devices are controlled by the master.
-    if iface_type not in ['slave']:
+    # Source 'interfaces' aren't brought up.
+    if iface_type not in ['slave', 'source']:
         return __salt__['cmd.run']('ifup {0}'.format(iface))
     return None
 
@@ -1722,7 +1793,7 @@ def get_routes(iface):
 
     .. code-block:: bash
 
-        salt '*' ip.get_interface eth0
+        salt '*' ip.get_routes eth0
     '''
 
     filename = os.path.join(_DEB_NETWORK_UP_DIR, 'route-{0}'.format(iface))
@@ -1747,16 +1818,33 @@ def apply_network_settings(**settings):
     if 'require_reboot' not in settings:
         settings['require_reboot'] = False
 
+    if 'apply_hostname' not in settings:
+        settings['apply_hostname'] = False
+
+    hostname_res = True
+    if settings['apply_hostname'] in _CONFIG_TRUE:
+        if 'hostname' in settings:
+            hostname_res = __salt__['network.mod_hostname'](settings['hostname'])
+        else:
+            log.warning(
+                'The network state sls is trying to apply hostname '
+                'changes but no hostname is defined.'
+            )
+            hostname_res = False
+
+    res = True
     if settings['require_reboot'] in _CONFIG_TRUE:
         log.warning(
             'The network state sls is requiring a reboot of the system to '
             'properly apply network configuration.'
         )
-        return True
+        res = True
     else:
         stop = __salt__['service.stop']('networking')
         time.sleep(2)
-        return stop and __salt__['service.start']('networking')
+        res = stop and __salt__['service.start']('networking')
+
+    return hostname_res and res
 
 
 def build_network_settings(**settings):
@@ -1809,11 +1897,13 @@ def build_network_settings(**settings):
 
     # Write hostname to /etc/hostname
     sline = opts['hostname'].split('.', 1)
-    hostname = '{0}\n' . format(sline[0])
+    opts['hostname'] = sline[0]
+    hostname = '{0}\n' . format(opts['hostname'])
     current_domainname = current_network_settings['domainname']
 
     # Only write the hostname if it has changed
     if not opts['hostname'] == current_network_settings['hostname']:
+        # TODO  replace wiht a call to network.mod_hostname instead
         _write_file_network(hostname, _DEB_HOSTNAME_FILE)
 
     new_domain = False
