@@ -6,6 +6,7 @@ from __future__ import absolute_import
 
 # Import python libs
 import os
+import re
 
 # Import salt libs
 import salt.utils
@@ -93,7 +94,7 @@ def get_all():
     ret = []
     service = _cmd()
     for svc in __salt__['cmd.run']('{0} ls all'.format(service)).splitlines():
-        ret.append(svc)
+            ret.append(svc)
     return sorted(ret)
 
 
@@ -110,7 +111,7 @@ def get_disabled():
     ret = []
     service = _cmd()
     for svc in __salt__['cmd.run']('{0} ls off'.format(service)).splitlines():
-        ret.append(svc)
+            ret.append(svc)
     return sorted(ret)
 
 
@@ -127,7 +128,7 @@ def get_enabled():
     ret = []
     service = _cmd()
     for svc in __salt__['cmd.run']('{0} ls on'.format(service)).splitlines():
-        ret.append(svc)
+            ret.append(svc)
     return sorted(ret)
 
 
@@ -224,14 +225,14 @@ def enable(name, **kwargs):
     stat_cmd = '{0} set {1} status on'.format(_cmd(), name)
     stat_retcode = __salt__['cmd.retcode'](stat_cmd)
 
-    set_retcode = None
+    flag_retcode = None
     # only (re)set flags for services that have an rc.d(8) script
     if os.path.exists('/etc/rc.d/{0}'.format(name)):
         flags = _get_flags(**kwargs)
-        set_cmd = '{0} set {1} flags {2}'.format(_cmd(), name, flags)
-        set_retcode = __salt__['cmd.retcode'](set_cmd)
+        flag_cmd = '{0} set {1} flags {2}'.format(_cmd(), name, flags)
+        flag_retcode = __salt__['cmd.retcode'](flag_cmd)
 
-    return not any([stat_retcode, set_retcode])
+    return not any([stat_retcode, flag_retcode])
 
 
 def disable(name, **kwargs):
