@@ -19,6 +19,10 @@ Connection module for Amazon S3
 
         s3.service_url: s3.amazonaws.com
 
+    A role_arn may also be specified in the configuration::
+
+        s3.role_arn: arn:aws:iam::111111111111:role/my-role-to-assume
+
     If a service_url is not specified, the default is s3.amazonaws.com. This
     may appear in various documentation as an "endpoint". A comprehensive list
     for Amazon S3 may be found at::
@@ -30,12 +34,18 @@ Connection module for Amazon S3
 
     SSL verification may also be turned off in the configuration:
 
-    s3.verify_ssl: False
+        s3.verify_ssl: False
 
     This is required if using S3 bucket names that contain a period, as
     these will not match Amazon's S3 wildcard certificates. Certificate
     verification is enabled by default.
 
+    AWS region may be specified in the configuration:
+
+        s3.location: eu-central-1
+
+    Default is us-east-1.
+
     This module should be usable to query other S3-like services, such as
     Eucalyptus.
 
@@ -61,7 +71,7 @@ def __virtual__():
 
 
 def delete(bucket, path=None, action=None, key=None, keyid=None,
-           service_url=None, verify_ssl=None):
+           service_url=None, verify_ssl=None, location=None, role_arn=None):
     '''
     Delete a bucket, or delete an object from a bucket.
 
@@ -73,8 +83,8 @@ def delete(bucket, path=None, action=None, key=None, keyid=None,
 
         salt myminion s3.delete mybucket remoteobject
     '''
-    key, keyid, service_url, verify_ssl = _get_key(key, keyid, service_url,
-                                                   verify_ssl)
+    key, keyid, service_url, verify_ssl, location, role_arn = _get_key(
+        key, keyid, service_url, verify_ssl, location, role_arn)
 
     return salt.utils.s3.query(method='DELETE',
                                bucket=bucket,
@@ -83,12 +93,14 @@ def delete(bucket, path=None, action=None, key=None, keyid=None,
                                key=key,
                                keyid=keyid,
                                service_url=service_url,
-                               verify_ssl=verify_ssl)
+                               verify_ssl=verify_ssl,
+                               location=location,
+                               role_arn=role_arn)
 
 
 def get(bucket=None, path=None, return_bin=False, action=None,
         local_file=None, key=None, keyid=None, service_url=None,
-        verify_ssl=None):
+        verify_ssl=None, location=None, role_arn=None):
     '''
     List the contents of a bucket, or return an object from a bucket. Set
     return_bin to True in order to retrieve an object wholesale. Otherwise,
@@ -140,8 +152,8 @@ def get(bucket=None, path=None, return_bin=False, action=None,
 
         salt myminion s3.get mybucket myfile.png action=acl
     '''
-    key, keyid, service_url, verify_ssl = _get_key(key, keyid, service_url,
-                                                   verify_ssl)
+    key, keyid, service_url, verify_ssl, location, role_arn = _get_key(
+        key, keyid, service_url, verify_ssl, location, role_arn)
 
     return salt.utils.s3.query(method='GET',
                                bucket=bucket,
@@ -152,11 +164,13 @@ def get(bucket=None, path=None, return_bin=False, action=None,
                                key=key,
                                keyid=keyid,
                                service_url=service_url,
-                               verify_ssl=verify_ssl)
+                               verify_ssl=verify_ssl,
+                               location=location,
+                               role_arn=role_arn)
 
 
 def head(bucket, path=None, key=None, keyid=None, service_url=None,
-         verify_ssl=None):
+         verify_ssl=None, location=None, role_arn=None):
     '''
     Return the metadata for a bucket, or an object in a bucket.
 
@@ -167,8 +181,8 @@ def head(bucket, path=None, key=None, keyid=None, service_url=None,
         salt myminion s3.head mybucket
         salt myminion s3.head mybucket myfile.png
     '''
-    key, keyid, service_url, verify_ssl = _get_key(key, keyid, service_url,
-                                                   verify_ssl)
+    key, keyid, service_url, verify_ssl, location, role_arn = _get_key(
+        key, keyid, service_url, verify_ssl, location, role_arn)
 
     return salt.utils.s3.query(method='HEAD',
                                bucket=bucket,
@@ -176,11 +190,15 @@ def head(bucket, path=None, key=None, keyid=None, service_url=None,
                                key=key,
                                keyid=keyid,
                                service_url=service_url,
-                               verify_ssl=verify_ssl)
+                               verify_ssl=verify_ssl,
+                               location=location,
+                               full_headers=True,
+                               role_arn=role_arn)
 
 
 def put(bucket, path=None, return_bin=False, action=None, local_file=None,
-        key=None, keyid=None, service_url=None, verify_ssl=None):
+        key=None, keyid=None, service_url=None, verify_ssl=None, location=None,
+        role_arn=None):
     '''
     Create a new bucket, or upload an object to a bucket.
 
@@ -196,8 +214,8 @@ def put(bucket, path=None, return_bin=False, action=None, local_file=None,
 
         salt myminion s3.put mybucket remotepath local_file=/path/to/file
     '''
-    key, keyid, service_url, verify_ssl = _get_key(key, keyid, service_url,
-                                                   verify_ssl)
+    key, keyid, service_url, verify_ssl, location, role_arn = _get_key(
+        key, keyid, service_url, verify_ssl, location, role_arn)
 
     return salt.utils.s3.query(method='PUT',
                                bucket=bucket,
@@ -208,10 +226,12 @@ def put(bucket, path=None, return_bin=False, action=None, local_file=None,
                                key=key,
                                keyid=keyid,
                                service_url=service_url,
-                               verify_ssl=verify_ssl)
+                               verify_ssl=verify_ssl,
+                               location=location,
+                               role_arn=role_arn)
 
 
-def _get_key(key, keyid, service_url, verify_ssl):
+def _get_key(key, keyid, service_url, verify_ssl, location, role_arn):
     '''
     Examine the keys, and populate as necessary
     '''
@@ -233,4 +253,10 @@ def _get_key(key, keyid, service_url, verify_ssl):
     if verify_ssl is None:
         verify_ssl = True
 
-    return key, keyid, service_url, verify_ssl
+    if location is None and __salt__['config.option']('s3.location') is not None:
+        location = __salt__['config.option']('s3.location')
+
+    if role_arn is None and __salt__['config.option']('s3.role_arn') is not None:
+        role_arn = __salt__['config.option']('s3.role_arn')
+
+    return key, keyid, service_url, verify_ssl, location, role_arn
